import os
import tempfile
import pytest

from unittest import TestCase
from unittest.mock import patch
from boto3 import Session
from botocore.exceptions import ClientError

from dli.client.exceptions import InsufficientPrivilegesException, S3FileDoesNotExist
from dli.client.s3 import Client, S3DatafileWrapper
from tests.localstack_helper import _fake_aws_credentials


session = Session(**_fake_aws_credentials())


@pytest.mark.integration
class S3ClientTestCase(TestCase):

    def setUp(self):
        self.target = Client(session)
        bucket = self.target.s3_resource.Bucket('bucket')
        bucket.create()

        def cleanup():
            bucket.objects.all().delete()
            bucket.delete()

        self.addCleanup(cleanup)

    def test_download_file_validates_path_exists_in_s3(self):
        with self.assertRaises(S3FileDoesNotExist):
            with tempfile.TemporaryDirectory() as dest:
                self.target.download_files_from_s3_path("s3://unknown/file", dest)

    def test_download_file_validates_destination_is_a_directory(self):
        self.target.s3fs.put(__file__, "s3://bucket/file")
        with self.assertRaises(NotADirectoryError):
            with tempfile.NamedTemporaryFile() as cm:
                self.target.download_files_from_s3_path("s3://bucket/file", cm.name)

    def test_download_file_creates_destination_directory_if_it_does_not_exist(self):
        self.target.s3fs.put(
            __file__,
            os.path.join("s3://bucket/location/", os.path.basename(__file__))
        )

        with tempfile.TemporaryDirectory() as dest:
            dest = os.path.join(dest, "dir1", "dir2")
            self.target.download_files_from_s3_path(
                "s3://bucket/location/test_s3_client.py",
                dest
            )

            self.assertTrue(os.path.exists(dest) and os.path.isdir(dest))
            self.assertTrue(os.path.exists(os.path.join(dest, "location", "test_s3_client.py")))

    def test_download_file_can_download_folders(self):
        self.target.s3fs.put(__file__, "s3://bucket/tdfcdf/file1.txt")
        self.target.s3fs.put(__file__, "s3://bucket/tdfcdf/subdir/file2.txt")
        self.target.s3fs.put(__file__, "s3://bucket/tdfcdf/subdir/subdir/file3.txt")

        with tempfile.TemporaryDirectory() as dest:
            self.target.download_files_from_s3_path("s3://bucket/tdfcdf", dest)

            self.assertTrue(os.path.exists(dest))
            self.assertTrue(os.path.exists(os.path.join(dest, "tdfcdf", "file1.txt")))
            self.assertTrue(os.path.exists(os.path.join(dest, "tdfcdf", "subdir", "file2.txt")))
            self.assertTrue(os.path.exists(os.path.join(dest, "tdfcdf", "subdir", "subdir", "file3.txt")))

    def test_download_file_can_download_single_file(self):
        self.target.s3fs.put(__file__, "s3://bucket/tdfcdf/file1.txt")

        with tempfile.TemporaryDirectory() as dest:
            self.target.download_files_from_s3_path("s3://bucket/tdfcdf/file1.txt", dest)

            self.assertTrue(os.path.exists(dest))
            self.assertTrue(os.path.exists(os.path.join(dest, "tdfcdf", "file1.txt")))


@pytest.mark.integration
class S3DatafileWrapperTestCase(TestCase):

    def setUp(self):
        self.client = Client(session)
        self.s3fs = self.client.s3fs
        bucket = self.client.s3_resource.Bucket('bucket')
        bucket.create()

        def cleanup():
            bucket.objects.all().delete()
            bucket.delete()

        self.addCleanup(cleanup)

    def test_files_returns_files_in_datafile(self):
        files = [
            "s3://bucket/tfrfid/a",
            "s3://bucket/tfrfid/b",
            "s3://bucket/tfrfid/c",
        ]

        for f in files:
            self.s3fs.touch(f)

        datafile = {
            "files": [{"path": f} for f in files]
        }

        target = S3DatafileWrapper(datafile, self.client)
        self.assertCountEqual(
            target.files,
            [f.replace("s3://", "") for f in files]
        )

    def test_files_will_recurse_directories(self):
        files = [
            "s3://bucket/tfwrd/a/aa",
            "s3://bucket/tfwrd/b/bb/bbb",
            "s3://bucket/tfwrd/c/cc/ccc/cccc1",
            "s3://bucket/tfwrd/c/cc/ccc/cccc2",
        ]

        # create sample files
        for f in files:
            self.s3fs.touch(f)

        datafile = {
            "files": [
                {'path': 's3://bucket/tfwrd/a'},
                {'path': 's3://bucket/tfwrd/b'},
                {'path': 's3://bucket/tfwrd/c'}
            ]
        }

        target = S3DatafileWrapper(datafile, self.client)

        self.assertCountEqual(
            target.files,
            [f.replace("s3://", "") for f in files]
        )

    def test_only_files_in_datafile_are_displayed(self):
        files = [
            "s3://bucket/tofidad/a",
            "s3://bucket/tofidad/b",
            "s3://bucket/tofidad/c",
            "s3://bucket/tofidad/d"
        ]

        # create sample files
        for f in files:
            self.s3fs.touch(f)

        datafile = {
            "files": [{"path": f} for f in files[0:1]]
        }
        target = S3DatafileWrapper(datafile, self.client)
        self.assertCountEqual(
            target.files,
            [f.replace("s3://", "") for f in files[0:1]]
        )

    def test_can_open_file(self):
        files = [
            "s3://bucket/tcof/a"
        ]
        # create sample files
        for f in files:
            self.s3fs.touch(f)
            with self.s3fs.open(f, mode="wb") as s3f:
                s3f.write(b"test 1")
                s3f.flush()

        datafile = {
            "files":  [{"path": f} for f in files]
        }

        target = S3DatafileWrapper(datafile, self.client)

        with target.open_file("bucket/tcof/a") as s3file:
            assert s3file is not None
            assert s3file.read() == b"test 1"

    def test_unknown_file_is_handled_gracefully(self):
        files = [
            "s3://bucket/tufihg/a"
        ]
        # create sample files
        for f in files:
            self.s3fs.touch(f)

        datafile = {
            "files": [{"path": f} for f in files]
        }

        target = S3DatafileWrapper(datafile, self.client)

        with self.assertRaises(S3FileDoesNotExist):
            res = target.open_file("bucket/unknown/file")


class S3UploadTestCase(TestCase):

    def setUp(self):
        self.patcher = patch('s3fs.S3FileSystem')
        self.mock_s3fs_instance = self.patcher.start().return_value
        client = Client(session)
        self.s3_client = client

        self.bucket = client.s3_resource.Bucket('s3_test')
        self.bucket.create()

        def cleanup():
            self.bucket.objects.all().delete()
            self.bucket.delete()

        self.addCleanup(cleanup)

        self.s3_location = 's3_test/temp/'

    def test_upload_files_to_s3_normal(self):
        # create some dummy files
        temp_file_list = [tempfile.NamedTemporaryFile() for i in range(3)]
        temp_file_path_list = [f.name for f in temp_file_list]

        expected_s3_location_list = [{
            "path": "s3://{}{}".format(self.s3_location, os.path.basename(file)),
            "size": 0
        } for file in temp_file_path_list]

        upload_result = self.s3_client.upload_files_to_s3(temp_file_path_list, self.s3_location)
        for temp_file in temp_file_list:
            temp_file.close()

        assert len(list(self.bucket.objects.all())) == 3
        self.assertCountEqual(upload_result, expected_s3_location_list)

    def test_upload_files_with_no_bucket_access_fails(self):
        error = ClientError(
            {'Error': {'Code': 'AccessDenied', 'Message': 'Access Denied'}},
            'NOOP'
        )

        with patch.object(self.s3_client, 's3_resource') as patched_s3_resource:
            patched_s3_resource.Bucket().Object().upload_file.side_effect = (
                error
            )

            with self.assertRaises(InsufficientPrivilegesException):
                self.s3_client.upload_files_to_s3([os.path.relpath(__file__)], self.s3_location)

    def tearDown(self):
        self.patcher.stop()



import unittest

import httpretty
import requests

from dli import __version__
from dli.client.dli_client import Session
from dli.client.exceptions import (
    DatalakeException,
    InsufficientPrivilegesException,
    UnAuthorisedAccessException
)


class SessionTestCase(unittest.TestCase):

    valid_token = "eyJ0eXAiOiJKV1QiLCJhbGciOiJIUzI1NiIsImlhdCI6MTUxMzMzMTUwOSwiZXhwIjo5NTEzMzM1MTA5fQ.e30.fjav6VOdXAnE6WgUwSxEbrNuiqQBSEalOd177F9UrCM"
    expired_token = "eyJ0eXAiOiJKV1QiLCJhbGciOiJIUzI1NiIsImlhdCI6MTUxMzMzMTUwOSwiZXhwIjoxNTEzMzM1MTA5fQ.e30.-vTTDvptpjPrG2PO1GBo1OOlBqvx6D_gPGnU7Tw-YII"

    def test_can_decode_valid_jwt_token(self):
        ctx = Session(
            "key",
            "root",
            None,
            self.valid_token
        )
        self.assertFalse(ctx.has_expired)

    def test_can_detect_token_is_expired(self):
        ctx = Session(
            "key",
            "root",
            None,
            self.expired_token
        )
        self.assertTrue(ctx.has_expired)

    def test_when_token_cant_be_decoded_then_we_assume_no_session_expiration(self):
        ctx = Session(
            "key",
            "root",
            None,
            "invalid.token"
        )
        self.assertFalse(ctx.has_expired)


class SessionRequestFactoryTestCase(unittest.TestCase):

    @httpretty.activate
    def test_response_403_raises_InsufficientPrivilegesException(self):
        response_text = 'Insufficient Privileges'
        httpretty.register_uri(httpretty.GET, 'http://dummy.com/test', status=403, body=response_text)

        session = Session(None, 'http://dummy.com', None, lambda: {})

        with self.assertRaises(InsufficientPrivilegesException):
            requests.Session().send(session.request_factory(method='GET', url='/test').prepare())

    @httpretty.activate
    def test_response_401_raises_UnAuthorisedAccessException(self):
        response_text = 'UnAuthorised Access'
        httpretty.register_uri(httpretty.GET, 'http://dummy.com/test', status=401, body=response_text)

        session = Session(None, 'http://dummy.com', None, lambda: {})

        with self.assertRaises(UnAuthorisedAccessException):
            requests.Session().send(session.request_factory(method='GET', url='/test').prepare())

    @httpretty.activate
    def test_response_500_raises_DatalakeException(self):
        response_text = 'Datalake server error'
        httpretty.register_uri(httpretty.GET, 'http://dummy.com/test', status=500, body=response_text)

        session = Session(None, 'http://dummy.com', None, lambda: {})

        with self.assertRaises(DatalakeException):
            requests.Session().send(session.request_factory(method='GET', url='/test').prepare())

    @httpretty.activate
    def test_sdk_version_is_included_in_header(self):
        httpretty.register_uri(httpretty.GET, 'http://dummy.com/test', status=200, body="response")
        session = Session(None, 'http://dummy.com', None, lambda: {})

        # issue a request
        requests.Session().send(session.request_factory(method='GET', url='/test').prepare())

        request = httpretty.last_request()
        self.assertTrue("X-Data-Lake-SDK-Version" in request.headers)
        self.assertEqual(request.headers["X-Data-Lake-SDK-Version"], str(__version__))



import pytest


def pytest_collection_modifyitems(config, items):
    skip_integration = config.getoption("--skip-integration")
    mark_skip_integration = pytest.mark.skip(reason="skipping integration test")
    for item in items:
        if skip_integration and "integration" in item.keywords:
            item.add_marker(mark_skip_integration)

def pytest_addoption(parser):
    parser.addoption(
        "--skip-integration", action="store_true", default=False,
        help="Skip integration tests."
    )



import datetime
import jwt
import requests
import logging

from collections.abc import Iterable
from http.cookiejar import CookiePolicy
from wrapt import ObjectProxy
from deprecated import deprecated
from http import HTTPStatus
from urllib.parse import urljoin, urlparse, parse_qs
from collections import defaultdict

from pypermedia import HypermediaClient
from requests_toolbelt.adapters import host_header_ssl

from dli import __version__
from dli.siren import siren_to_entity
from dli.client.auth import get_auth_key
from dli.client.components.auto_reg_metadata import AutoRegMetadata
from dli.client.components.collection import Collection
from dli.client.components.datafile import Datafile
from dli.client.components.dataset import Dataset, Dictionary
from dli.client.components.me import Me
from dli.client.components.package import Package
from dli.client.components.search import Search
from dli.client.exceptions import (
    DatalakeException, InsufficientPrivilegesException, InvalidPayloadException,
    UnAuthorisedAccessException, CatalogueEntityNotFoundException
)

from dli.siren import PatchedSirenBuilder

logger = logging.getLogger(__name__)


class DliClient:
    """
    Definition of a client. This client mixes in utility functions for
    manipulating collections, packages, datasets and datafiles.
    """
    def __init__(self, api_key, api_root, host=None):
        self.api_key = api_key
        self.api_root = api_root
        self.host = host
        self._session = self._new_session()
        self._components = [
            component(client=self)
            for component in (
                AutoRegMetadata,
                Collection,
                Datafile,
                Dataset,
                Me,
                Package,
                Search,
                Dictionary,
            )
        ]
        self.methods = {
            method.__name__: method
            for method in (
                getattr(component, attr)
                for component in self._components
                for attr in dir(component)
                if not attr.startswith('_')
            )
            if callable(method)
        }

    def _new_session(self):
        return Session(
            self.api_key,
            self.api_root,
            self.host,
        )

    def __getattr__(self, item):
        if item in self.methods:
            return self.methods[item]
        else:
            raise AttributeError(
                f"'{type(self)}' object has no attribute '{item}'"
            )

    @property
    def session(self):
        # if the session expired, then reauth
        # and create a new context
        if self._session.has_expired:
            self._session = self._new_session()
        return self._session

    @property
    def ctx(self):
        """ Alias to self.session for backward compatibility.
        """
        return self.session


class Response(ObjectProxy):

    def __init__(self, wrapped, builder, *args, **kwargs):
        super(Response, self).__init__(wrapped, *args, **kwargs)
        self.builder = builder

    @property
    def is_siren(self):
        return self.request.headers.get(
            'Content-Type', ''
        ).startswith('application/vnd.siren+json')

    def to_siren(self):
        # Pypermedias terminology, not mine
        python_object = self.builder._construct_entity(
            self.json()
        ).as_python_object()

        # Keep the response availible
        python_object._raw_response = self

        return python_object

    def to_many_siren(self, relation):
        return [
            siren_to_entity(c) for c in
            self.to_siren().get_entities(rel=relation)
        ]


class BlockAll(CookiePolicy):
    return_ok = set_ok = domain_return_ok = path_return_ok = lambda self, *args, **kwargs: False
    netscape = True
    rfc2965 = hide_cookie2 = False


class Session(requests.Session):

    def __init__(self, api_key, api_root, host, auth_key=None):
        super().__init__()
        self.api_key = api_key
        self.api_root = api_root
        self.host = host
        self.auth_key = auth_key or get_auth_key(api_key, api_root, host)

        if host and urlparse(api_root).scheme == "https":
            self.mount(
                'https://',
                host_header_ssl.HostHeaderSSLAdapter()
            )

        # Preserve backwards compat
        self.request_session = self

        self.verify = True
        self.s3_keys = {}
        self.token_expires_on = self.get_expiration_date(self.auth_key)
        self.siren_builder = PatchedSirenBuilder(
            verify=self.verify,
            request_factory=self.request_factory
        )
        # Don't allow cookies to be set.
        # The new API will reject requests with both a cookie
        # and a auth header (as there's no predictiable crediential to choose).
        #
        # However the old API, once authenticate using a Bearer token will
        # as a side effect of a request return a oidc_id_token which matches
        # the auth header. This is ignored.
        self.cookies.set_policy(BlockAll())

    def request_factory(self, method=None, url=None, headers=None, files=None, data=None,
                        params=None, auth=None, cookies=None, hooks=None, json=None):
        # relative uri? make it absolute.
        if not urlparse(url).netloc:
            url = urljoin(str(self.api_root), str(url))     # python 2/3 nonsense

        if isinstance(data, Iterable):
            json = data
            data = None

        headers = headers or {}
        # populate headers
        if urlparse(url).path.startswith('/__api/'):
            # Only the old API uses this. We want to be able to
            # override this.
            headers['Content-Type'] = "application/vnd.siren+json"

        headers["X-Data-Lake-SDK-Version"] = str(__version__)
        headers.update(self.get_header_with_auth())

        # if a host has been provided, then we need to set it on the header
        if self.host:
            headers['Host'] = self.host

        if not hooks:
            hooks = {}
        if 'response' not in hooks:
            hooks['response'] = []
        hooks['response'] = self.make_hook()

        return requests.Request(
            method=method,
            url=url,
            headers=headers,
            files=files,
            data=data,
            params=params,
            auth=auth,
            hooks=hooks,
            json=json
        )

    def request(self, method, url, cert=None, verify=None,
                stream=None, proxies=None, timeout=None, allow_redirects=True, **kwargs):
        """Constructs a :class:`PreparedRequest <PreparedRequest>` for
        transmission and returns it. The :class:`PreparedRequest` has settings
        merged from the :class:`Request <Request>` instance and those of the
        :class:`Session`.
        :param request: :class:`Request` instance to prepare with this
            session's settings.
        :rtype: requests.PreparedRequest
        """
        # appreciate this is very ugly, attempting to maintain backwards compat
        req = self.request_factory(
            method=method.upper(),
            url=url,
            **kwargs
        )

        prep = self.prepare_request(req)

        proxies = proxies or {}

        settings = self.merge_environment_settings(
            prep.url, proxies, stream, verify, cert
        )

        # Send the request.
        send_kwargs = {
            'timeout': timeout,
            'allow_redirects': allow_redirects,
        }
        send_kwargs.update(settings)
        resp = self.send(prep, **send_kwargs)

        return resp

    @staticmethod
    @deprecated(reason='No longer used for anything - private method')
    def create_request_session(root, host):
        # build the requests sessions that pypermedia will use
        # to submit requests
        session = requests.Session()

        # when no dns is available and the user is using an ip address
        # to reach the catalogue
        # we need to force the cert validation to be against
        # the host header, and not the host in the uri
        # (we only do this if the scheme of the root is https)
        if host and urlparse(root).scheme == "https":
            session.mount(
                'https://',
                host_header_ssl.HostHeaderSSLAdapter()
            )

        return session

    @staticmethod
    def get_expiration_date(token):
        # use a default_timeout if the token can't be decoded
        # until the proper endpoint is added on the catalog
        default_timeout = (
            datetime.datetime.utcnow() +
            datetime.timedelta(minutes=55)
        )

        try:
            # get the expiration from the jwt auth token
            decoded_token = jwt.get_unverified_header(token)
            if 'exp' not in decoded_token:
                return default_timeout

            return datetime.datetime.fromtimestamp(
                decoded_token['exp']
            )

        except Exception:
            return default_timeout

    @property
    def has_expired(self):
        # by default we don't want to fail if we could not decode the token
        # so if the ``token_expiry`` is undefined we assume the session
        # is valid
        if not self.token_expires_on:
            return False
        return datetime.datetime.utcnow() > self.token_expires_on

    def get_header_with_auth(self):
        auth_header = "Bearer {}".format(self.auth_key)
        return {"Authorization": auth_header}

    def uri_with_root(self, relative_path):
        return "{}/{}".format(self.api_root, relative_path)

    def get_root_siren(self):
        return HypermediaClient.connect(
            self.api_root,
            request_factory=self.request_factory,
            session=self,
            builder=PatchedSirenBuilder,
            verify=self.verify
        )

    def make_hook(self):
        return self._response_hook

    def _response_hook(self, response, *args, **kwargs):
        # Appologies for the ugly code. The startswith siren check
        # is to make this onlly relevant to the old API.
        response = Response(response, self.siren_builder)

        if not response.ok:
            exceptions = defaultdict(
                lambda: DatalakeException,
                {HTTPStatus.BAD_REQUEST: InvalidPayloadException,
                 HTTPStatus.UNPROCESSABLE_ENTITY: InvalidPayloadException,
                 HTTPStatus.UNAUTHORIZED: UnAuthorisedAccessException,
                 HTTPStatus.FORBIDDEN: InsufficientPrivilegesException,
                 HTTPStatus.NOT_FOUND: CatalogueEntityNotFoundException}
            )

            try:
                message = response.json()
            except ValueError:
                message = response.text

            raise exceptions[response.status_code](
                message=message,
                params=parse_qs(urlparse(response.request.url).query),
                response=response,
            )

        return response



from http import HTTPStatus

import requests
from requests_toolbelt.adapters import host_header_ssl


class AuthenticationFailure(Exception):
    """
    An exception wrapping an authentication failure response. If the response
    had a payload, that payload is reported as the exception message, otherwise
    a generic error message is returned.
    """
    GENERIC_ERROR_MESSAGE = (
        'API key is not valid or has expired.'
        ' Please generate a new key in the Catalogue UI and try again.'
    )

    def __init__(self, response=None, message=None):
        self.response = response
        self.message = message

    def __str__(self):
        if self.response and self.response.text:
            return self.response.text
        elif self.message:
            return self.message
        return self.GENERIC_ERROR_MESSAGE


def get_auth_key(api_key, api_root, host=None):
    key = api_key
    auth_header = "Bearer {}".format(key)
    start_session_url = "{}/start-session".format(api_root)

    session = requests.Session()
    headers = {"Authorization": auth_header}

    # if a host has been provided, then we need to set it on the header
    # and activate the certificate check against the header itself
    # rather than the ip address
    if host:
        session.mount('https://', host_header_ssl.HostHeaderSSLAdapter())
        headers["Host"] = host

    r = session.post(start_session_url, headers=headers)
    if r.status_code == HTTPStatus.FORBIDDEN:
        api_key_url = api_root.replace('__api', 'dashboard/my-api-keys')
        error_message = "%s Catalog API Key URL: %s" % (
        AuthenticationFailure.GENERIC_ERROR_MESSAGE, api_key_url)
        raise AuthenticationFailure(message=error_message)
    elif r.status_code != HTTPStatus.OK:
        raise AuthenticationFailure(response=r)

    return r.text



import requests
import re

from collections import namedtuple
from pypermedia.client import SirenBuilder
from pypermedia.siren import SirenEntity, _create_action_fn

import six

# regex to convert from camelCase to snake_case
_reg = re.compile(r'(?!^)(?<!_)([A-Z])')


def siren_to_dict(o):
    return {
        attr: getattr(o, attr)
        for attr in dir(o)
        if attr[:1] != '_' and not callable(getattr(o, attr))
    }


def siren_to_entity(o):
    """
    Helper method that converts a siren entity into a namedtuple
    """
    def value_to_entity(v):
        return siren_to_entity(v) if isinstance(v, dict) else v

    # pypermedia does not do recursive translation
    # so we might get a mix of objects or dicts in here
    attrs = siren_to_dict(o) if not isinstance(o, dict) else o
    attrs = {
        to_snake_case(key): value_to_entity(value)
        for key, value in six.iteritems(attrs)
    }
    return namedtuple(o.__class__.__name__, sorted(attrs))(**attrs)


def to_snake_case(s):
    # need to sanitise the string as in some cases the key might look like
    # 'Data Access' or 'Data Sensitivity'
    return _reg.sub(r'_\1', s.replace(' ', '_')).lower()


class PatchedSirenBuilder(SirenBuilder):
    def from_api_response(self, response):
        """Overriding to provide raw response data"""
        if not response.headers.get(
            'Content-Type', ''
        ).startswith('application/vnd.siren+json'):
            return None

        return super().from_api_response(response)

    def _construct_entity(self, entity_dict):
        """
        We need to patch the actions as there is no ``radio`` support
        on the current pypermedia version.

        To avoid code duplication, this function will attempt to call the parent
        and replace the created actions with our custom ones.
        """
        # pypermedia does not like any custom attributes
        # not even those in the spec
        for action in entity_dict.get("actions", []):
            if "allowed" in action:
                del action["allowed"]

        entity = super(PatchedSirenBuilder, self)._construct_entity(entity_dict)

        return PatchedSirenEntity(
            classnames=entity.classnames,
            properties=entity.properties,
            actions=entity.actions,
            links=entity.links,
            entities=entity.entities,
            rel=entity.rel,
            verify=entity.verify,
            request_factory=entity.request_factory
        )


def _patched_make_request(self, _session=None, verify=None, **kwfields):
    """ The purpose of this is to override the `SirenAction.make_request` in
        pypermedia, which ignores the passed `verify` argument and just uses
        the one defined on self, while the one passed into the function gets
        swallowed into `kwfields` and sent in the request.
    """
    if verify is None:
        verify = self.verify
    s = _session or requests.Session()
    return s.send(self.as_request(**kwfields), verify=verify)


class PatchedSirenEntity(SirenEntity):

    def as_python_object(self):
        ModelClass = type(str(self.get_primary_classname()), (), self.properties)

        siren_builder = PatchedSirenBuilder(verify=self.verify, request_factory=self.request_factory)
        # add actions as methods
        for action in self.actions:
            # patch make_request to resolve InsecureRequestWaring problems (see DL-1961)
            action.make_request = _patched_make_request.__get__(action, type(action))
            method_name = SirenEntity._create_python_method_name(action.name)
            method_def = _create_action_fn(action, siren_builder)
            setattr(ModelClass, method_name, method_def)

        # add links as methods
        for link in self.links:
            for rel in link.rel:
                method_name = SirenEntity._create_python_method_name(rel)
                siren_builder = PatchedSirenBuilder(verify=self.verify, request_factory=self.request_factory)
                method_def = _create_action_fn(link, siren_builder)

                setattr(ModelClass, method_name, method_def)

        def get_entity(obj, rel):
            matching_entities = self.get_entities(rel) or []
            for x in matching_entities:
                yield x.as_python_object()

        setattr(ModelClass, 'get_entities', get_entity)

        return ModelClass()



from dli.client.components import SirenComponent
from dli.client.utils import ensure_count_is_valid


class Search(SirenComponent):
    """
    Catalogue search functions.
    """

    class urls:
        search_root = '/__api/search/'
        search_packages = '/__api/search/packages/'
        search_datasets = '/__api/search/datasets/'
        search_collections = '/__api/search/collections/'

    def _search_my_entities(self, entity, search_func, term, count):
        ensure_count_is_valid(count)

        # replicating UI behavior, for empty term we want an empty search
        if term is None or term == "":
            return []

        result = search_func(query=term, page_size=count)
        return result.get_entities(rel=entity)

    def search(self, term, count=100):
        """
        Search across all catalogue entities given a particular set of keywords.

        :param str term: The search term.
        :param int count: Optional. The amount of results to be returned.

        :returns: A list of Catalogue entities
        :rtype: list[collections.namedtuple]

        - **Sample**

        .. code-block:: python

                automotive_catalogue_entities = client.search(
                    term="Automotive",
                    count=100
                )

        """
        ensure_count_is_valid(count)
        return self.client.session.get(
            self.urls.search_root, params={'query': term, 'page_size': count}
        ).to_many_siren('')

    def search_packages(self, term, count=100):
        """
        Search across packages in the catalogue given a particular set of keywords.

        :param str term: The search term.
        :param int count: Optional. The amount of results to be returned. Defaults to 100.

        :returns: A list of packages
        :rtype: list[collections.namedtuple]

        - **Sample**

        .. code-block:: python

                automotive_packages = client.search_packages(
                    term="Automotive",
                    count=100
                )

        """
        ensure_count_is_valid(count)
        return self.client.session.get(
            self.urls.search_packages, params={'query': term, 'page_size': count}
        ).to_many_siren('package')


    def search_collections(self, term, count=100):
        """
        Search across collections in the catalogue given a particular set of keywords.

        :param str term: The search term.
        :param int count: Optional. The amount of results to be returned. Defaults to 100.

        :returns: A list of collections
        :rtype: list[collections.namedtuple]

        - **Sample**

        .. code-block:: python

                automotive_collections = client.search_collections(
                    term="Automotive",
                    count=100
                )

        """
        ensure_count_is_valid(count)
        return self.client.session.get(
            self.urls.search_collections, params={'query': term, 'page_size': count}
        ).to_many_siren('collection')

    def search_datasets(self, term, count=100):
        """
        Search across datasets in the catalogue given a particular set of keywords.

        :param str term: The search term.
        :param int count: Optional. The amount of results to be returned. Defaults to 100.

        :returns: A list of datasets
        :rtype: list[dataset.namedtuple]

        - **Sample**

        .. code-block:: python

                results = client.search_datasets(
                    term="CDS",
                    count=100
                )

        """
        ensure_count_is_valid(count)
        return self.client.session.get(
            self.urls.search_datasets, params={'query': term, 'page_size': count}
        ).to_many_siren('dataset')



import contextlib

from collections.abc import Mapping


class AttributesDict(Mapping):

    def __init__(self, *args, **kwargs):
        # recurisvely provide the rather silly attribute
        # access
        data = {}

        for arg in args:
            data.update(arg)

        data.update(**kwargs)

        for key, value in data.items():
            if isinstance(value, Mapping):
                self.__dict__[key] = AttributesDict(value)
            else:
                self.__dict__[key] = value

    def __getitem__(self, key):
        return self.__dict__[key]

    def __iter__(self):
        return iter(self.__dict__)

    def __len__(self):
        return len(self.__dict__)

    def _asdict(self, *args, **kwargs):
        return self

    def __repr__(self):
        attributes = ' '.join([
            '{}={}'.format(k, v) for k,v in self.__dict__.items()
        ])

        return "{}({})".format(self.__class__.__name__, attributes)


class SampleData:
    def __init__(self, parent):
        self._parent = parent
        self._client = parent._client.client
        self._urls = parent._client.urls

    def schema(self):
        """
        Returns the schema data and first rows of sample data.

        :returns: attributes dictionary
        """
        response = self._client.session.get(
            self._urls.v2_sample_data_schema.format(id=self._parent.id)
        )

        return AttributesDict(**response.json()['data']['attributes'])

    @contextlib.contextmanager
    def file(self):
        """
        Provides a file like object containing sample data.

        Example usage:

        .. code-block:: python

            dataset = self.client.get_dataset(dataset_id)
            with dataset.sample_data.file() as f:
                dataframe = pandas.read_csv(f)
        
        """
        response = self._client.session.get(
            self._urls.v2_sample_data_file.format(id=self._parent.id),
            stream=True
        )
        yield response.raw
        response.close()


class DatasetModel(AttributesDict):

    @property
    def sample_data(self):
        return SampleData(self)

    @property
    def id(self):
        return self.dataset_id

    def __init__(self, json, client=None):
        self._client = client

        location = json['attributes'].pop('location')

        if 's3' in location:
            location = location['s3']
        else:
            location = location['other']

        super().__init__(
            json['attributes'],
            dataset_id=json['id'],
            location=location
        )


class SchemaModel(AttributesDict):

    @property
    def id(self):
        return self.dictionary_id

    @property
    def schema_id(self):
        return self.id

    def __init__(self, json):
        super().__init__(
            json['attributes'],
            dictionary_id=json['id'],
        )



import json


class DatalakeException(Exception):
    def __init__(self, entity=None, params=None, message=None, response=None):
        if not params and not message:
            message = "%s not found" % entity

        self.params = params
        self.message = (
            message or 'entity with params `{}` not found'.format(params)
        )
        self.entity = entity
        self.response = response

        super(DatalakeException, self).__init__(
            self.message, entity,
            params, message
        )

    def __str__(self):
        return (
            f'\n{self.response.request.method} {self.response.request.url}\n\n'
            f'{json.dumps(self.response.json(), indent=4)}'
        )


class CatalogueEntityNotFoundException(DatalakeException):
    pass


class InvalidPayloadException(DatalakeException):
    pass


class UnAuthorisedAccessException(DatalakeException):
    pass


class InsufficientPrivilegesException(DatalakeException):
    message = 'Insufficient privileges to perform this action'


class S3FileDoesNotExist(DatalakeException):
    def __init__(self, file_path):
        self.message = (
            "Either file at path `%s` does not exist / Potential issue with the bucket policy."
            "Please reach out to Datalake Tech Data Ops user for resolution." % file_path
        )

        super(S3FileDoesNotExist, self).__init__(self.message)


class DownloadFailed(DatalakeException):
    pass


class NoAccountSpecified(DatalakeException):

    def __init__(self, accounts):
        self.accounts = accounts
        self.message = (
            "Unable to default the account for access_manager_id, tech_data_ops_id and/or manager_id "
            "due to multiple accounts being attached to this API key. "
            "Your accounts are: %s" % [(a.id, a.name) for a in accounts]
        )
        super(NoAccountSpecified, self).__init__(self.message)
